import datetime

import click
import colorama
import termcolor

import netdumplings
from netdumplings import DumplingDriver, DumplingEater
from netdumplings._shared import HUB_HOST, HUB_OUT_PORT

from netdumplings.console._shared import CLICK_CONTEXT_SETTINGS, printable_dumpling


class PrinterEater(DumplingEater):
    """
    A dumpling eater which displays dumpling information to the terminal as it
    arrives from ``nd-hub``.
    """
    def __init__(
            self,
            kitchens=None,
            interval_dumplings=True,
            packet_dumplings=True,
            payload=True,
            color=True,
            **kwargs):

        super().__init__(**kwargs)

        self._kitchens = kitchens
        self._interval_dumplings = interval_dumplings
        self._packet_dumplings = packet_dumplings
        self._payload = payload
        self._color = color

        if color:
            colorama.init()  # Needed for Windows console.

    async def on_connect(self, hub_uri, websocket):
        """
        Called when the connection to ``nd-hub`` has been created.

        :param hub_uri: The ``nd-hub`` websocket URI.
        :param websocket: The websocket object used for talking to nd-hub
            (websockets.WebSocketClientProtocol).
        """
        print('Connected to nd-hub at {0}'.format(hub_uri))
        print('Waiting for dumplings...\n')

    async def on_dumpling(self, dumpling):
        """
        Called when a new dumpling is received from ``nd-hub``. Prints the
        dumpling summary and payload.

        :param dumpling: The received dumpling.
        """
        driver = dumpling.driver

        should_print_dumpling = (
            (driver == DumplingDriver.interval and self._interval_dumplings) or
            (driver == DumplingDriver.packet and self._packet_dumplings)
        ) and (
            self._kitchens is None or dumpling.kitchen in self._kitchens
        )

        if not should_print_dumpling:
            return

        dumpling_creation_time = (
            datetime.datetime.fromtimestamp(dumpling.creation_time).isoformat()
        )

        dumpling_chef = (
            termcolor.colored(dumpling.chef_name, attrs=['bold'])
            if self._color else dumpling.chef_name
        )

        dumpling_kitchen = (
            termcolor.colored(dumpling.kitchen, attrs=['bold'])
            if self._color else dumpling.kitchen
        )

        summary = '{} [{:8s}] {} from {}'.format(
            dumpling_creation_time,
            'packet' if driver == DumplingDriver.packet else 'interval',
            dumpling_chef,
            dumpling_kitchen,
        )

        print(summary)

        if self._payload:
            print('\n{}\n'.format(
                printable_dumpling(dumpling.payload, colorize=self._color)
            ))

    async def on_connection_lost(self, e):
        """
        Called when the ``nd-hub`` connection is lost.

        :param e: The exception thrown during the connection close.
        """
        print('\nLost connection to nd-hub: {}'.format(e))


# -----------------------------------------------------------------------------

@click.command(
    context_settings=CLICK_CONTEXT_SETTINGS,
)
@click.option(
    '--hub', '-h',
    help='Address where nd-hub is sending dumplings from.',
    metavar='HOST:PORT',
    default='{}:{}'.format(HUB_HOST, HUB_OUT_PORT),
    show_default=True,
)
@click.option(
    '--chef', '-c',
    help='Restrict dumplings to those made by this chef. Multiple can be '
         'specified. Displays all chefs by default.',
    metavar='CHEF_NAME',
    multiple=True,
)
@click.option(
    '--kitchen', '-k',
    help='Restrict dumplings to those emitted by this kitchen. Multiple can '
         'be specified. Displays all kitchens by default.',
    metavar='KITCHEN_NAME',
    multiple=True,
)
@click.option(
    '--eater-name', '-n',
    help='Dumpling eater name for this tool when connecting to nd-hub.',
    default='printereater',
    metavar='EATER_NAME',
    show_default=True,
)
@click.option(
    '--interval-dumplings / --no-interval-dumplings',
    help='Print interval dumplings.',
    default=True,
    show_default=True,
)
@click.option(
    '--packet-dumplings / --no-packet-dumplings',
    help='Print packet dumplings.',
    default=True,
    show_default=True,
)
@click.option(
    '--payload / --no-payload',
    help='Print dumpling payload.',
    default=True,
    show_default=True,
)
@click.option(
    '--color / --no-color',
    help='Print color output.',
    default=True,
    show_default=True,
)
@click.version_option(version=netdumplings.__version__)
def print_cli(hub, chef, kitchen, eater_name, interval_dumplings,
              packet_dumplings, payload, color):
    """
    A dumpling eater.

    Connects to nd-hub (the dumpling hub) and prints information on all
    received dumplings.
    """
    eater = PrinterEater(
        kitchens=kitchen if kitchen else None,
        interval_dumplings=interval_dumplings,
        packet_dumplings=packet_dumplings,
        payload=payload,
        color=color,
        name=eater_name,
        hub=hub,
        chef_filter=chef if chef else None,
    )

    eater.run()


if __name__ == '__main__':
    print_cli()



from .dumpling import Dumpling, DumplingDriver
from .dumplingchef import DumplingChef
from .dumplingeater import DumplingEater
from .exceptions import (
    InvalidDumpling, InvalidDumplingPayload, NetDumplingsError,
)
from .dumplinghub import DumplingHub
from .dumplingkitchen import DumplingKitchen
from ._version import __version__

# Workaround to avoid F401 "imported but unused" linter errors.
(
    Dumpling,
    DumplingDriver,
    DumplingChef,
    DumplingEater,
    DumplingHub,
    DumplingKitchen,
    InvalidDumpling,
    InvalidDumplingPayload,
    NetDumplingsError,
    __version__,
)



from enum import Enum
import json
import time
from typing import Any, Optional, Union

from .dumplingchef import DumplingChef
from .exceptions import InvalidDumpling, InvalidDumplingPayload


class DumplingDriver(Enum):
    """
    The event driving the creation of a :class:`Dumpling`.

    ``DumplingDriver.packet``
        a network packet being received by a :class:`DumplingChef`.

    ``DumplingDriver.interval``
        a regular time interval poke being received by a :class:`DumplingChef`.
    """
    packet = 1
    interval = 2


class Dumpling:
    """
    Represents a single Dumpling.

    Dumplings are usually created automatically by a dumpling kitchen from the
    payload returned by a :class:`~DumplingChef`.

    A Dumpling exposes the following attributes:

    * ``chef`` - the :class:`DumplingChef` which is responsible for the
      dumpling payload
    * ``chef_name`` - the name of the DumplingChef
    * ``kitchen`` - the name of the kitchen which created the Dumpling
    * ``driver`` - the :class:`DumplingDriver` for the dumpling
    * ``creation_time`` - when the dumpling was created (epoch milliseconds)
    * ``payload`` - the dumpling payload

    :param chef: The chef which created the dumpling payload (usually a
        :class:`DumplingChef` instance, but can be a string).
    :param driver: The event type that drove the dumpling to be created.
    :param creation_time: Dumpling creation time (epoch milliseconds). Defaults
        to current time.
    :param payload: The dumpling payload information. Can be anything (usually
        a dict) which is JSON-serializable.
    """
    def __init__(
            self,
            *,
            chef: Union[DumplingChef, str],
            driver: DumplingDriver = DumplingDriver.packet,
            creation_time: Optional[float] = None,
            payload: Any) -> None:

        self.chef = chef

        try:
            self.chef_name = chef.name
            self.kitchen = chef.kitchen
        except AttributeError:
            self.chef_name = chef
            self.kitchen = None

        self.driver = driver
        self.creation_time = (
            time.time() if creation_time is None else creation_time
        )
        self.payload = payload

    def __repr__(self):
        if self.driver == DumplingDriver.packet:
            driver = 'DumplingDriver.packet'
        elif self.driver == DumplingDriver.interval:
            driver = 'DumplingDriver.interval'
        else:
            driver = repr(self.driver)

        payload = (None if self.payload is None
                   else '<{}>'.format(type(self.payload).__name__))

        return (
            '{}('
            'chef={}, '
            'driver={}, '
            'creation_time={}, '
            'payload={})'.format(
                type(self).__name__,
                repr(self.chef),
                driver,
                self.creation_time,
                payload,
            )
        )

    @classmethod
    def from_json(cls, json_dumpling: str):
        """
        A Dumpling factory which creates a Dumpling from a given
        ``json_dumpling`` string. The given ``json_dumpling`` is expected to be
        a dumpling which has already been JSON-serialized (presumably by a
        dumpling kitchen).

        :param json_dumpling: JSON string to create the Dumpling from.
        :return: A :class:`Dumpling` instance.
        :raise: :class:`InvalidDumpling` if ``json_dumpling`` could not be
            successfully converted into a Dumpling.
        """
        try:
            dumpling_dict = json.loads(json_dumpling)
        except (json.decoder.JSONDecodeError, TypeError, ValueError) as e:
            raise InvalidDumpling(
                'Could not interpret dumpling JSON: {}'.format(e)
            )

        metadata = dumpling_dict['metadata']

        try:
            if metadata['driver'] == 'packet':
                driver = DumplingDriver.packet
            elif metadata['driver'] == 'interval':
                driver = DumplingDriver.interval
            else:
                raise InvalidDumpling(
                    "Dumpling driver was not 'packet' or 'interval'"
                )

            dumpling = cls(
                chef=metadata['chef'],
                driver=driver,
                creation_time=metadata['creation_time'],
                payload=dumpling_dict['payload'],
            )

            dumpling.kitchen = metadata['kitchen']
            return dumpling
        except KeyError as e:
            raise InvalidDumpling(
                'Dumpling JSON was missing key: {}'.format(e)
            )

    def to_json(self) -> str:
        """
        Creates a JSON-serialized dumpling string from the Dumpling.

        :return: A JSON string representation of the Dumpling.
        :raise: :class:`InvalidDumplingPayload` if the Dumpling payload cannot
            be JSON-serialized.
        """
        dumpling = {
            'metadata': {
                'chef': self.chef_name,
                'kitchen': self.kitchen.name if self.kitchen else None,
                'creation_time': self.creation_time,
                'driver': self.driver.name,
            },
            'payload': self.payload,
        }

        try:
            dumpling_serialized = json.dumps(dumpling)
        except TypeError as e:
            raise InvalidDumplingPayload(
                'Could not create dumpling: {}'.format(e)
            )

        return dumpling_serialized



import asyncio
import json
import logging
import signal
from typing import Callable, List, Optional
import websockets

from .dumpling import Dumpling
from .exceptions import InvalidDumpling

from ._shared import ND_CLOSE_MSGS, HUB_HOST, HUB_OUT_PORT


class DumplingEater:
    """
    Base helper class for Python-based dumpling eaters.

    Connects to ``nd-hub`` and listens for any dumplings made by the provided
    ``chef_filter`` (or all chefs if ``chef_filter`` is ``None``). Can be
    given ``async`` callables for any of the following events:

    ``on_connect(websocket_uri, websocket_obj)``
        invoked when the connection to ``nd-hub`` is made

    ``on_dumpling(dumpling)``
        invoked whenever a dumpling is emitted from ``nd-hub``

    ``on_connection_lost(e)``
        invoked when the connection to ``nd-hub`` is closed

    **The above callables must be** ``async def`` **methods**.

    :param name: Name of the dumpling eater. Is ideally unique per eater.
    :param hub: Address where ``nd-hub`` is sending dumplings from.
    :param chef_filter: List of chef names whose dumplings this eater wants to
        receive. ``None`` means get all chefs' dumplings.
    :param on_connect: Called when connection to ``nd-hub`` is made. Is passed
        two parameters: the ``nd-hub`` websocket URI (string) and websocket
        object (:class:`websockets.client.WebSocketClientProtocol`).
    :param on_dumpling: Called whenever a dumpling is received. Is passed the
        dumpling as a Python dict.
    :param on_connection_lost: Called when connection to ``nd-hub`` is lost. Is
        passed the associated exception object.
    """
    def __init__(
            self,
            name: str = 'nameless_eater',
            hub: str ='{}:{}'.format(HUB_HOST, HUB_OUT_PORT),
            *,
            chef_filter: Optional[List[str]] = None,
            on_connect: Optional[Callable] = None,
            on_dumpling: Optional[Callable] = None,
            on_connection_lost: Optional[Callable] = None) -> None:

        self.name = name
        self.chef_filter = chef_filter
        self.hub = hub
        self.hub_ws = "ws://{0}".format(hub)

        # Configure handlers. If we're not provided with handlers then we
        # fall back on the default handlers or the handlers provided by a
        # subclass.
        self.on_connect = (
            on_connect if on_connect is not None else self.on_connect
        )
        self.on_dumpling = (
            on_dumpling if on_dumpling is not None else self.on_dumpling
        )
        self.on_connection_lost = (
            on_connection_lost if on_connection_lost is not None
            else self.on_connection_lost
        )

        self._was_connected = False
        self._logger_name = "{}.{}".format(__name__, self.name)
        self.logger = logging.getLogger(self._logger_name)

    def __repr__(self):
        def handler_string(attr):
            # We can't use 'repr(self.handler)' for callables because it causes
            # an infinite loop as the repr of the handler includes the repr of
            # the handler (etc). So we replace handler reprs with
            # '<callable: name>'.
            return (
                '<callable: {}>'.format(attr.__name__) if callable(attr)
                else repr(attr)
            )

        return (
            '{}('
            'name={}, '
            'hub={}, '
            'chef_filter={}, '
            'on_connect={}, '
            'on_dumpling={}, '
            'on_connection_lost={})'.format(
                type(self).__name__,
                repr(self.name),
                repr(self.hub),
                repr(self.chef_filter),
                handler_string(self.on_connect),
                handler_string(self.on_dumpling),
                handler_string(self.on_connection_lost),
            )
        )

    async def _grab_dumplings(self, dumpling_count=None):
        """
        Receives all dumplings from the hub and looks for any dumplings which
        were created by the chef(s) we're interested in. All those dumplings
        are then passed to the on_dumpling handler (after being converted from
        their JSON form back into a Dumpling instance).

        :param dumpling_count: Number of dumplings to eat. ``None`` means eat
            forever.
        """
        dumplings_eaten = 0

        websocket = await websockets.client.connect(self.hub_ws)
        self._was_connected = True

        self.logger.info("{0}: Connected to dumpling hub at {1}".format(
            self.name, self.hub_ws))

        try:
            # Announce ourselves to the dumpling hub.
            await websocket.send(json.dumps({'eater_name': self.name}))

            if self.on_connect:
                await self.on_connect(self.hub_ws, websocket)

            while True:
                # Eat a single dumpling.
                dumpling_json = await websocket.recv()

                # Create a Dumpling from the JSON received over the websocket.
                # Note that invalid dumplings will probably be stripped out by
                # the hub already.
                try:
                    dumpling = Dumpling.from_json(dumpling_json)
                except InvalidDumpling as e:
                    self.logger.error("{0}: Invalid dumpling: {1}".format(
                        self.name, e))
                    continue

                self.logger.debug("{0}: Received dumpling from {1}".format(
                    self.name, dumpling.chef_name))

                # Call the on_dumpling handler if this dumpling is from a
                # chef that we've registered interest in.
                if (self.chef_filter is None or
                        dumpling.chef_name in self.chef_filter):
                    self.logger.debug(
                        "{0}: Calling dumpling handler {1}".format(
                            self.name, self.on_dumpling))

                    dumplings_eaten += 1
                    await self.on_dumpling(dumpling)

                # Stop eating dumplings if we've reached our threshold.
                if dumpling_count is not None and \
                        dumplings_eaten >= dumpling_count:
                    await websocket.close(*ND_CLOSE_MSGS['eater_full'])
                    break
        except asyncio.CancelledError:
            self.logger.warning(
                f"\n{self.name}: Connection to dumpling hub cancelled; "
                f"closing..."
            )

            try:
                await websocket.close(*ND_CLOSE_MSGS['conn_cancelled'])
            except websockets.exceptions.InvalidState:
                pass
        except websockets.exceptions.ConnectionClosed as e:
            self.logger.warning(
                "{}: Lost connection to dumpling hub: {}".format(self.name, e)
            )

            if self.on_connection_lost:
                await self.on_connection_lost(e)

    @staticmethod
    def _interrupt_handler():
        """
        Signal handler. Cancels all running async tasks.
        """
        tasks = asyncio.Task.all_tasks()
        for task in tasks:
            task.cancel()

    def run(self, dumpling_count=None):
        """
        Run the dumpling eater.

        This will block until the desired ``dumpling_count`` is met.

        :param dumpling_count: Number of dumplings to eat. ``None`` means eat
            forever.
        """
        self.logger.info("{0}: Running dumpling eater".format(self.name))

        if not callable(self.on_dumpling):
            self.logger.error(
                "{0}: on_dumpling handler is not callable".format(self.name))
            return

        self.logger.debug("{0}: Looking for dumpling hub at {1}".format(
            self.name, self.hub_ws))
        self.logger.debug("{0}: Chefs: {1}".format(
            self.name,
            ", ".join(self.chef_filter) if self.chef_filter else 'all')
        )

        try:
            asyncio.run(self._grab_dumplings(dumpling_count))
        except OSError as e:
            self.logger.warning(
                "{0}: There was a problem with the dumpling hub connection. "
                "Is nd-hub available?".format(self.name))
            self.logger.warning("{0}: {1}".format(self.name, e))
        finally:
            if self._was_connected:
                self.logger.warning(
                    "{0}: Done eating dumplings.".format(self.name))

    async def on_connect(self, websocket_uri, websocket_obj):
        """
        Default on_connect handler.

        This will be used if an ``on_connect`` handler is not provided during
        instantiation, and if a handler is not provided by a DumplingEater
        subclass.

        Only logs an warning-level log entry.
        """
        self.logger.warning(
            '{}: No on_connect handler specified; ignoring '
            'connection.'.format(self.name)
        )

    async def on_dumpling(self, dumpling):
        """
        Default on_dumpling handler.

        This will be used if an ``on_dumpling`` handler is not provided during
        instantiation, and if a handler is not provided by a DumplingEater
        subclass.

        Only logs an warning-level log entry.
        """
        self.logger.warning(
            '{}: No on_dumpling handler specified; ignoring '
            'dumpling.'.format(self.name)
        )

    async def on_connection_lost(self, e):
        """
        Default on_connection_lost handler.

        This will be used if an ``on_connection_lost`` handler is not provided
        during instantiation, and if a handler is not provided by a
        DumplingEater subclass.

        Only logs an warning-level log entry.
        """
        self.logger.warning(
            '{}: No on_connection_lost handler specified; ignoring '
            'connection loss.'.format(self.name)
        )



import asyncio
import datetime
import json
import logging
import websockets
from websockets.exceptions import ConnectionClosed

from .dumpling import Dumpling, DumplingDriver
from .exceptions import InvalidDumpling, NetDumplingsError

from ._shared import (
    validate_dumpling, HUB_HOST, HUB_IN_PORT, HUB_OUT_PORT, HUB_STATUS_FREQ,
)


class DumplingHub:
    """
    Implements a dumpling hub.

    A dumpling hub is two websocket servers: one receives dumplings from any
    number of running ``nd-sniff`` scripts; and the other sends those dumplings
    to any number of dumpling eaters. The hub also makes its own dumplings
    which describe its own system status which are also sent to all the
    dumpling eaters at regular intervals.

    ``nd-hub`` is a simple wrapper around ``DumplingHub``.

    :param address: Address the hub is running on.
    :param in_port: Port used to receive connections from `nd-sniff`.
    :param out_port: Port used to receive connections from `dumpling eaters`.
    :param status_freq: Frequency (in secs) to send system status dumplings.
    """
    def __init__(
            self,
            address: str = HUB_HOST,
            in_port: int = HUB_IN_PORT,
            out_port: int = HUB_OUT_PORT,
            status_freq: int = HUB_STATUS_FREQ,
    ) -> None:

        self.address = address
        self.in_port = in_port
        self.out_port = out_port
        self.status_freq = status_freq

        # Maintain a dictionary of all connected kitchens and eaters.  The
        # key is the websocket and the value is a dictionary of information
        # on the kitchen/eater.
        self._dumpling_eaters = {}
        self._dumpling_kitchens = {}

        self._start_time = datetime.datetime.now()

        self._system_stats = {
            'dumplings_in': 0,
            'dumplings_out': 0
        }

        self._logger = logging.getLogger(__name__)

    def __repr__(self):
        return (
            '{}('
            'address={}, '
            'in_port={}, '
            'out_port={}, '
            'status_freq={})'.format(
                type(self).__name__,
                repr(self.address),
                repr(self.in_port),
                repr(self.out_port),
                repr(self.status_freq),
            )
        )

    def _get_system_status(self):
        """
        Generates current system status information.

        :return: System status information.
        """
        uptime = (datetime.datetime.now() - self._start_time).total_seconds()

        system_status = {
            'total_dumplings_in': self._system_stats['dumplings_in'],
            'total_dumplings_out': self._system_stats['dumplings_out'],
            'server_uptime': uptime,
            'dumpling_kitchen_count': len(self._dumpling_kitchens),
            'dumpling_eater_count': len(self._dumpling_eaters),
            'dumpling_kitchens':
                [self._dumpling_kitchens[kitchen]['metadata']
                 for kitchen in self._dumpling_kitchens],
            'dumpling_eaters':
                [self._dumpling_eaters[eater]['metadata']
                 for eater in self._dumpling_eaters]
        }

        return system_status

    async def _grab_dumplings(self, websocket, path):
        """
        A coroutine for grabbing dumplings from a single instance of
        ``nd-sniff``. A single instance of this coroutine exists for each
        connected ``nd-sniff`` and is invoked via
        :meth:`websockets.server.serve`.

        :param websocket: A :class:`websockets.server.WebSocketServerProtocol`.
        :param path: Websocket request URI.
        """
        host = websocket.remote_address[0]
        port = websocket.remote_address[1]

        # Retain some information on this dumpling kitchen.
        kitchen_json = await websocket.recv()
        kitchen = {
            'metadata': {
                'info_from_kitchen': json.loads(kitchen_json),
                'info_from_hub': {
                    'host': host,
                    'port': port
                }
            },
            'websocket': websocket
        }

        self._dumpling_kitchens[websocket] = kitchen
        kitchen_name = kitchen['metadata']['info_from_kitchen']['kitchen_name']

        self._logger.info(
            "Received dumpling kitchen connection from {0} at {1}:{2}".format(
                kitchen_name, host, port))

        try:
            while True:
                dumpling_json = await websocket.recv()

                # Validate the dumpling.
                try:
                    dumpling = validate_dumpling(dumpling_json)
                except InvalidDumpling as e:
                    self._logger.error(
                        "Received invalid dumpling: {0}; kitchen: {1}".format(
                            e,
                            json.dumps(
                                kitchen['metadata']['info_from_kitchen']
                            )
                        ))
                    continue

                self._system_stats['dumplings_in'] += 1

                chef = dumpling['metadata']['chef']
                self._logger.debug(
                    "Received {} dumpling from {} at {}:{}; {} bytes".format(
                        chef, kitchen_name, host, port, len(dumpling_json)))

                # Send this dumpling to all the eager dumpling eaters.
                for eater in self._dumpling_eaters:
                    await self._dumpling_eaters[eater]['queue'].put(
                        dumpling_json
                    )
        except ConnectionClosed as e:
            self._logger.info(
                "Dumpling kitchen {0} connection closed: {1}".format(
                    kitchen_name, e))
            del self._dumpling_kitchens[websocket]

    async def _emit_dumplings(self, websocket, path):
        """
        A coroutine for sending all received dumplings to a single connected
        dumpling eater over a websocket connection. A single instance of this
        coroutine exists for each eater and is invoked via
        :meth:`websockets.server.serve`.

        :param websocket: A :class:`websockets.server.WebSocketServerProtocol`.
        :param path: Websocket request URI.
        """
        host = websocket.remote_address[0]
        port = websocket.remote_address[1]

        # Retain some information on this dumpling eater.
        eater_json = await websocket.recv()
        eater = {
            'metadata': {
                'info_from_eater': json.loads(eater_json),
                'info_from_hub': {
                    'host': host,
                    'port': port
                }
            },
            'websocket': websocket,
            'queue': asyncio.Queue()
        }

        self._dumpling_eaters[websocket] = eater
        eater_name = eater['metadata']['info_from_eater']['eater_name']

        self._logger.info(
            "Received dumpling eater connection from {0} at {1}:{2}".format(
                eater_name, host, port))

        # Each dumpling eater has their own queue.  These queues receive all
        # the fresh new dumplings received by each instance of the
        # _grab_dumplings coroutine.
        dumpling_queue = eater['queue']

        try:
            while True:
                dumpling = await dumpling_queue.get()
                dumpling_obj = json.loads(dumpling)
                chef = dumpling_obj['metadata']['chef']

                self._logger.debug(
                    "Sending {0} dumpling to {1} at {2}:{3}; {4} bytes".format(
                        chef, eater_name, host, port, len(dumpling)))

                await websocket.send(dumpling)
                self._system_stats['dumplings_out'] += 1
        except ConnectionClosed as e:
            self._logger.info(
                "Dumpling eater {0} connection closed: {1}".format(
                    eater_name, e))
            del self._dumpling_eaters[websocket]

    async def _announce_system_status(self):
        """
        Sends system status (as a dumpling) to all connected dumpling eaters.

        :param freq_secs: Frequency (in seconds) of status announcements.
        """
        while True:
            # We create our own system status dumplings (rather than going
            # through a chef+kitchen pair).
            status_dumpling = Dumpling(
                chef='SystemStatusChef', driver=DumplingDriver.interval,
                payload=self._get_system_status())

            status_dumpling_json = status_dumpling.to_json()

            for eater in self._dumpling_eaters:
                await self._dumpling_eaters[eater]['queue'].put(
                    status_dumpling_json
                )

            await asyncio.sleep(self.status_freq)

    def run(self):
        """
        Run the dumpling hub.

        Starts two websocket servers: one to receive dumplings from zero or
        more instances of ``nd-sniff``; and another to send those dumplings to
        zero or more dumpling eaters. Also creates its own dumplings at regular
        intervals to send system status information to all connected dumpling
        eaters.
        """
        dumpling_in_server = \
            websockets.serve(self._grab_dumplings, self.address, self.in_port)
        dumpling_out_server = \
            websockets.serve(self._emit_dumplings, self.address, self.out_port)

        loop = asyncio.get_event_loop()

        try:
            srv_in = loop.run_until_complete(dumpling_in_server)
            srv_out = loop.run_until_complete(dumpling_out_server)
        except OSError as e:
            raise NetDumplingsError(
                "Cannot instantiate dumpling hub: {0}".format(e))

        status_task = asyncio.ensure_future(self._announce_system_status())

        self._logger.info("Dumpling hub initiated; waiting for connections")

        in_uri = "ws://{0}:{1}".format(self.address, self.in_port)
        out_uri = "ws://{0}:{1}".format(self.address, self.out_port)
        self._logger.info(
            "Dumplings in: {0}  out: {1}".format(in_uri, out_uri)
        )

        try:
            loop.run_forever()
        except KeyboardInterrupt:
            self._logger.warning(
                "Caught keyboard interrupt; attempting graceful shutdown...")
        except asyncio.CancelledError:
            pass
        finally:
            srv_in.close()
            srv_out.close()
            loop.run_until_complete(srv_in.wait_closed())
            loop.run_until_complete(srv_out.wait_closed())
            self._logger.info("Dumpling hub signing off.  Thanks!")



import sys

import click

import netdumplings
from netdumplings.exceptions import NetDumplingsError
from netdumplings._shared import (
    configure_logging, HUB_HOST, HUB_IN_PORT, HUB_OUT_PORT, HUB_STATUS_FREQ,
)

from netdumplings.console._shared import CLICK_CONTEXT_SETTINGS


@click.command(
    context_settings=CLICK_CONTEXT_SETTINGS,
)
@click.option(
    '--address', '-a',
    help='Address where nd-hub will send dumplings from.',
    metavar='HOSTNAME',
    default=HUB_HOST,
    show_default=True,
)
@click.option(
    '--in-port', '-i',
    help='Port to receive incoming dumplings from.',
    metavar='PORT',
    type=click.INT,
    default=HUB_IN_PORT,
    show_default=True,
)
@click.option(
    '--out-port', '-o',
    help='Port to send outgoing dumplings on.',
    metavar='PORT',
    type=click.INT,
    default=HUB_OUT_PORT,
    show_default=True,
)
@click.option(
    '--status-freq', '-f',
    help='Frequency (in seconds) to send status dumplings.',
    metavar='SECONDS',
    type=click.INT,
    default=HUB_STATUS_FREQ,
    show_default=True,
)
@click.version_option(version=netdumplings.__version__)
def hub_cli(address, in_port, out_port, status_freq):
    """
    The dumpling hub.

    Sends dumplings received from all kitchens (usually any running instances
    of nd-sniff) to all dumpling eaters. All kitchens and eaters need to
    connect to the nd-hub --in-port or --out-port respectively.
    """
    configure_logging()

    dumpling_hub = netdumplings.DumplingHub(
        address=address,
        in_port=in_port,
        out_port=out_port,
        status_freq=status_freq,
    )

    try:
        dumpling_hub.run()
    except NetDumplingsError as e:
        print('hub error: {0}'.format(e))
        sys.exit(1)


if __name__ == '__main__':
    hub_cli()



from .hub import hub_cli
from .hubdetails import hubdetails_cli
from .hubstatus import hubstatus_cli
from .print import print_cli
from .sniff import sniff_cli


(
    hub_cli,
    hubdetails_cli,
    hubstatus_cli,
    print_cli,
    sniff_cli,
)



import asyncio
import importlib.util
import json
import logging
import os.path
import queue
import sys
import multiprocessing
from typing import Dict, List, Optional, Union

import click
import websockets

import netdumplings
from netdumplings._shared import (
    configure_logging, ND_CLOSE_MSGS, HUB_HOST, HUB_IN_PORT,
)

from netdumplings.console._shared import CLICK_CONTEXT_SETTINGS


def network_sniffer(
        kitchen_name: str,
        interface: str,
        chefs: Union[List[str], bool],
        chef_modules: List[str],
        valid_chefs: Dict,
        sniffer_filter: str,
        chef_poke_interval: int,
        dumpling_queue: multiprocessing.Queue,
):
    """
    The main network sniffing management function, responsible for:

    * Instantiating a dumpling kitchen (which does the actual sniffing) and
      providing it with a queue to put chef-created dumplings on
    * Instantiating the dumpling chefs and registering them with the kitchen
    * Running the kitchen

    This function is intended to be invoked as a Python process.

    A dumpling chef will only be instantiated if its ``assignable_to_kitchen``
    class attribute is ``True``.

    :param kitchen_name: Name of the sniffer kitchen.
    :param interface: Network interface to sniff (``all`` sniffs all
        interfaces).
    :param chefs: List of chefs to send packets to. Used for display only.
    :param chef_modules: List of Python module names in which to find chefs.
        Used for display only.
    :param valid_chefs: Dict of module+chef combinations we plan on importing.
    :param sniffer_filter: PCAP-compliant sniffer filter.
    :param chef_poke_interval: Interval (in secs) to poke chefs.
    :param dumpling_queue: Queue to pass to the kitchen to put dumplings on.
    """
    configure_logging()
    log = logging.getLogger('netdumplings.console.sniff')
    log.info("{0}: Starting network sniffer process".format(kitchen_name))
    log.info("{0}: Interface: {1}".format(kitchen_name, interface))
    log.info("{0}: Requested chefs: {1}".format(kitchen_name,
             "all" if chefs is True else ", ".join(chefs)))
    log.info("{0}: Chef modules: {1}".format(
        kitchen_name, ", ".join(chef_modules)))
    log.info("{0}: Filter: {1}".format(kitchen_name,
             "<all packets>" if not sniffer_filter else sniffer_filter))
    log.info("{0}: Chef poke interval (secs): {1}".format(
        kitchen_name, chef_poke_interval))

    sniffer_kitchen = netdumplings.DumplingKitchen(
        dumpling_queue=dumpling_queue,
        name=kitchen_name,
        interface=interface,
        sniffer_filter=sniffer_filter,
        chef_poke_interval=chef_poke_interval,
    )

    # Instantiate all the valid DumplingChef classes and register them with
    # the kitchen.
    for chef_module in valid_chefs:
        chef_class_names = valid_chefs[chef_module]

        if os.path.isfile(chef_module):
            spec = importlib.util.spec_from_file_location('chefs', chef_module)
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)
        else:
            # TODO: Investigate replacing __import__ with
            #   importlib.import_module
            mod = __import__(chef_module, fromlist=chef_class_names)

        for chef_class_name in chef_class_names:
            log.info("{0}: Registering {1}.{2} with kitchen".format(
                kitchen_name, chef_module, chef_class_name))
            klass = getattr(mod, chef_class_name)
            klass(kitchen=sniffer_kitchen)

    sniffer_kitchen.run()


async def send_dumplings_from_queue_to_hub(
        kitchen_name: str,
        hub: str,
        dumpling_queue: multiprocessing.Queue,
        kitchen_info: dict,
        log: logging.Logger,
):
    """
    Grabs dumplings from the dumpling queue and sends them to ``nd-hub``.

    :param kitchen_name: The name of the kitchen.
    :param hub: The address where ``nd-hub`` is receiving dumplings.
    :param dumpling_queue: Queue to grab dumplings from.
    :param kitchen_info: Dict describing the kitchen.
    :param log: Logger.
    """
    hub_ws = 'ws://{0}'.format(hub)

    log.info("{0}: Connecting to the dumpling hub at {1}".format(
        kitchen_name, hub_ws)
    )

    try:
        websocket = await websockets.client.connect(hub_ws)
    except OSError as e:
        log.error(
            "{0}: There was a problem with the dumpling hub connection. "
            "Is nd-hub available?".format(kitchen_name))
        log.error("{0}: {1}".format(kitchen_name, e))
        return

    try:
        # Register our kitchen information with the dumpling hub.
        await websocket.send(json.dumps(kitchen_info))

        # Send dumplings to the hub when they come in from the chefs.
        while True:
            # This is a bit hacky. We have a multiprocessing queue to get from,
            # but we're running in a coroutine. The get() blocks, which I think
            # is inferfering with websockets' ability to manage its heartbeat
            # with the hub. This only seems to affect Windows. The workaround
            # implemented here is to put a 1-second timeout on the queue get,
            # ignore empty gets, and await asyncio.sleep() which appears to
            # allow the run loop to continue (presumably allowing the keepalives
            # to work).
            try:
                dumpling = dumpling_queue.get(timeout=1)
                await websocket.send(dumpling)
            except queue.Empty:
                pass

            await asyncio.sleep(0)
    except asyncio.CancelledError:
        log.warning(
            "{0}: Connection to dumpling hub cancelled; closing...".format(
                kitchen_name))
        try:
            await websocket.close(*ND_CLOSE_MSGS['conn_cancelled'])
        except websockets.exceptions.InvalidState:
            pass
    except websockets.exceptions.ConnectionClosed as e:
        log.warning("{0}: Lost connection to dumpling hub: {1}".format(
            kitchen_name, e))
    except OSError as e:
        log.exception(
            "{0}: Error talking to dumpling hub: {1}".format(kitchen_name, e)
        )


def dumpling_emitter(
        kitchen_name: str,
        hub: str,
        dumpling_queue: multiprocessing.Queue,
        kitchen_info: Dict,
):
    """
    Starts an async event loop to manage funneling dumplings from the queue to
    the dumpling hub.

    This function is intended to be invoked as a Python process.

    :param kitchen_name: The name of the kitchen that the dumplings will be
        coming from.
    :param hub: The address where ``nd-hub`` is receiving dumplings.
    :param dumpling_queue: Queue to get dumplings from.
    :param kitchen_info: Information on the kitchen.
    """
    configure_logging()
    log = logging.getLogger('netdumplings.console.sniff')
    log.info("{0}: Starting dumpling emitter process".format(kitchen_name))

    try:
        asyncio.run(
            send_dumplings_from_queue_to_hub(
                kitchen_name, hub, dumpling_queue, kitchen_info, log
            )
        )
    except KeyboardInterrupt:
        log.warning(f"Keyboard interrupt detected")


def list_chefs(chef_modules: Optional[List[str]] = None):
    """
    Lists all the chef classes (subclassed from :class:`DumplingChef`) found in
    the given list of ``chef_modules``.

    :param chef_modules: Python module names to look for chefs in.
    """
    chef_info = netdumplings.DumplingKitchen.get_chefs_in_modules(chef_modules)

    print()
    for chef_module in sorted(chef_info):
        print("{0}".format(chef_module))
        import_error = chef_info[chef_module]['import_error']

        if not import_error:
            for chef_class in chef_info[chef_module]['chef_classes']:
                print("  {0}".format(chef_class))
        else:
            print("  error importing module: {0}".format(import_error))

        print()


def get_valid_chefs(
        kitchen_name: str,
        chef_modules: List[str],
        chefs_requested: Union[List[str], bool],
        log: logging.Logger,
) -> Dict:
    """
    Retrieves the names of all valid DumplingChef subclasses for later
    instantiation. Valid chefs are all the classes in ``chef_modules`` which
    subclass DumplingChef and are included in our list of ``chefs_requested``.
    They also need to have their ``assignable_to_kitchen`` attribute set to
    True.

    :param kitchen_name: Kitchen name (for logging purposes).
    :param chef_modules: List of modules to look for chefs in.
    :param chefs_requested: List of requested chef names (True means all chefs
        are requested).
    :param log: Logger to log to.
    :return: Dict of valid DumpingChef subclasses. Keys are the Python module
        names and the values are a list of valid chef class names in each
        module.
    """
    valid_chefs = {}
    chef_info = netdumplings.DumplingKitchen.get_chefs_in_modules(chef_modules)
    # TODO: chefs_seen could be a set.
    chefs_seen = []

    # Find all the valid chefs.
    for chef_module in chef_info:
        import_error = chef_info[chef_module]['import_error']
        if import_error:
            log.error('Problem with {}: {}'.format(chef_module, import_error))
            continue

        chef_class_names = chef_info[chef_module]['chef_classes']
        is_py_file = chef_info[chef_module]['is_py_file']

        if is_py_file:
            spec = importlib.util.spec_from_file_location('chefs', chef_module)
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)
        else:
            # TODO: Investigate replacing __import__ with
            #   importlib.import_module
            mod = __import__(chef_module, fromlist=chef_class_names)

        for chef_class_name in chef_class_names:
            chefs_seen.append(chef_class_name)
            klass = getattr(mod, chef_class_name)
            if not klass.assignable_to_kitchen:
                log.warning("{0}: Chef {1} is marked as unassignable".format(
                    kitchen_name, chef_class_name))
                continue

            # A chefs_requested value of True means all chefs.
            if chefs_requested is True or chef_class_name in chefs_requested:
                try:
                    valid_chefs[chef_module].append(chef_class_name)
                except KeyError:
                    valid_chefs[chef_module] = [chef_class_name]

    # Warn about any requested chefs which were not found.
    if chefs_requested is not True:
        for chef_not_found in [chef for chef in chefs_requested
                               if chef not in chefs_seen]:
            log.warning("{0}: Chef {1} not found".format(
                kitchen_name, chef_not_found))

    return valid_chefs


# -----------------------------------------------------------------------------

@click.command(
    context_settings=CLICK_CONTEXT_SETTINGS,
)
@click.option(
    '--kitchen-name', '-n',
    help='Dumpling kitchen name to assign to the sniffer',
    metavar='KITCHEN_NAME',
    default='default_kitchen',
    show_default=True,
)
@click.option(
    '--hub', '-h',
    help='Address where nd-hub is receiving dumplings.',
    metavar='HOST:PORT',
    default='{}:{}'.format(HUB_HOST, HUB_IN_PORT),
    show_default=True,
)
@click.option(
    '--interface', '-i',
    help='Network interface to sniff.',
    metavar='INTERFACE',
    default='all',
    show_default=True,
)
@click.option(
    '--filter', '-f', 'pkt_filter',
    help='PCAP-style sniffer packet filter.',
    metavar='PCAP_FILTER',
    default='tcp or udp or arp',
    show_default=True,
)
@click.option(
    '--chef-module', '-m',
    help='Python module containing chef implementations. Can be module.name '
         'or /path/to/file.py. Multiple can be specified.',
    metavar='PYTHON_MODULE',
    default=['netdumplings.dumplingchefs'],
    show_default=True,
    multiple=True,
)
@click.option(
    '--chef', '-c',
    help='Chef (as found in a --chef-module) to deliver packets to. Multiple '
         'can be specified. Default is to send packets to all chefs.',
    metavar='CHEF_NAME',
    multiple=True,
)
@click.option(
    '--chef-list', '-l',
    help='List all available chefs (as found in the given --chef-module '
         'Python modules) and exit.',
    is_flag=True,
    default=False,
)
@click.option(
    '--poke-interval', '-p',
    help='Interval (in seconds) to poke chefs instructing them to send their '
         'interval dumplings.',
    metavar='SECONDS',
    type=click.FLOAT,
    default=5.0,
    show_default=True,
)
@click.version_option(version=netdumplings.__version__)
def sniff_cli(kitchen_name, hub, interface, pkt_filter, chef_module, chef,
              chef_list, poke_interval):
    """
    A dumpling sniffer kitchen.

    Sniffs network packets matching the given PCAP-style filter and sends them
    to chefs for processing into dumplings. Dumplings are then sent to nd-hub
    for distribution to the dumpling eaters.

    This tool likely needs to be run as root, or as an Administrator user.
    """
    # NOTE: Since the --chef-module and --chef flags can be specified multiple
    #   times, the associated 'chef_module' and 'chef' parameters are tuples of
    #   zero or more modules/chefs respectively.
    chef = True if len(chef) == 0 else chef

    # Display the chef list and exit if that's all the user wanted.
    if chef_list:
        list_chefs(chef_module)
        sys.exit(0)

    # now do the following:
    #
    # * Create a queue for a network-sniffing kitchen process to put dumplings
    #   onto
    # * Start a kitchen process, which will be putting dumplings onto the queue
    # * Start a dumpling emitter process which takes dumplings from the queue
    #   and sends them to nd-hub over a websocket

    configure_logging()
    logger = logging.getLogger('netdumplings.console.sniff')
    logger.info("Initializing sniffer...")

    # A queue for passing dumplings from the sniffer kitchen to the
    # dumpling-emitter process.
    dumpling_emitter_queue = multiprocessing.Queue()

    # Determine what chefs we'll be sending packets to.
    valid_chefs = get_valid_chefs(kitchen_name, chef_module, chef, logger)

    if not valid_chefs:
        logger.error('{}: No valid chefs found. Not starting sniffer.'.format(
            kitchen_name
        ))
        sys.exit(1)

    # Generate list of module.class names for all the seemingly-valid chefs
    # we'll be instantiating.  This is for use in the status dumplings.
    valid_chef_list = []
    for chef_module_name in sorted(valid_chefs.keys()):
        for chef_class_name in sorted(valid_chefs[chef_module_name]):
            valid_chef_list.append('{}.{}'.format(
                chef_module_name, chef_class_name)
            )

    # Start the sniffer kitchen and dumpling-emitter processes.
    sniffer_process = multiprocessing.Process(
        target=network_sniffer,
        args=(
            kitchen_name, interface, chef, chef_module, valid_chefs,
            pkt_filter, poke_interval, dumpling_emitter_queue,
        )
    )

    kitchen_info = {
        'kitchen_name': kitchen_name,
        'interface': interface,
        'filter': pkt_filter,
        'chefs': valid_chef_list,
        'poke_interval': poke_interval,
    }

    dumpling_emitter_process = multiprocessing.Process(
        target=dumpling_emitter,
        args=(kitchen_name, hub, dumpling_emitter_queue, kitchen_info),
        daemon=True,
    )

    sniffer_process.start()
    dumpling_emitter_process.start()

    try:
        while True:
            sniffer_process.join(0.5)
            dumpling_emitter_process.join(0.5)

            if not sniffer_process.is_alive():
                logger.error(
                    f"{kitchen_name}: Network sniffer process died; exiting."
                )
                break

            if not dumpling_emitter_process.is_alive():
                logger.error(
                    f"{kitchen_name}: Dumpling emitter process died; exiting."
                )
                break
    except KeyboardInterrupt:
        logger.warning(f"{kitchen_name}: Caught keyboard interrupt; exiting.")

    for process in multiprocessing.active_children():
        process.terminate()


if __name__ == '__main__':
    sniff_cli()



class NetDumplingsError(Exception):
    """
    Base exception for all netdumplings errors.
    """


class InvalidDumpling(NetDumplingsError):
    """
    A JSON-serialized dumpling does not appear to be valid.
    """


class InvalidDumplingPayload(NetDumplingsError):
    """
    A dumpling payload is invalid (probably because it is not
    JSON-serializable).
    """



import time

from netdumplings import DumplingChef


class DNSLookupChef(DumplingChef):
    """
    Makes dumplings which describe DNS activity.  Sends per-packet dumplings
    for individual DNS (Domain Name System) lookups; and poke-interval
    dumplings which describe the hosts lookups seen so far with per-host lookup
    counts and timestamp of last lookup.

    Dumpling payload examples:

    Per DNS lookup: ::

        {
            "lookup": {
                "hostname": "srirachamadness.com",
                "when": 1499040017.811247
            }
        }

    Per poke-interval: ::

        {
            "lookups_seen": {
                "srirachamadness.com": {
                    "count": 28,
                    "latest": 1499040017.811247
                },
                "www.fleegle.com": {
                    "count": 1,
                    "latest": 1499045642.57563
                },
                "floople.com": {
                    "count": 7,
                    "latest": 1499043343.104648
                }
            }
        }
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lookups_seen = {}

    def packet_handler(self, packet):
        """
        Processes a packet from nd-sniff.  Makes a dumpling summarizing the
        contents of each each valid DNS lookup.

        :param packet: Packet from nd-sniff.
        """
        if not packet.haslayer('DNS'):
            return

        dns_query = packet.getlayer('DNS')
        query = dns_query.fields['qd']

        if query is None:
            return

        hostname = query.qname.decode('utf-8')

        if hostname.endswith('.'):
            hostname = hostname[:-1]

        now_millis = time.time()

        try:
            self.lookups_seen[hostname]['count'] += 1
            self.lookups_seen[hostname]['latest'] = now_millis
        except KeyError:
            self.lookups_seen[hostname] = {
                'count': 1,
                'latest': now_millis,
            }

        payload = {
            'lookup': {
                'hostname': hostname,
                'when': now_millis
            }
        }

        return payload

    def interval_handler(self, interval=None):
        """
        Makes a dumpling at regular intervals which summarizes all the host
        lookups seen so far along with the count and latest lookup time for
        each host.
        """
        payload = {
            'lookups_seen': self.lookups_seen
        }

        return payload



from collections import defaultdict

from netdumplings import DumplingChef


class PacketCountChef(DumplingChef):
    """
    Makes dumplings which describe the total number of packets seen so far per
    network layer.  Sends one dumpling per poke interval.

    Dumpling payload example: ::

        {
            "packet_counts": {
                "DNS": 9678,
                "Ethernet": 2659740,
                "IP": 2367035,
                "IPv6": 292705,
                "NTP": 28,
                "Raw": 2420309,
                "TCP": 535050,
                "UDP": 2124690
            }
        }
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.packet_counts = defaultdict(int)
        self.poke_count = 0

    def packet_handler(self, packet):
        """
        Processes a packet from nd-sniff.  Adds 1 to the layer count for each
        layer found in the packet.  This does not make any dumplings.

        :param packet: Packet from nd-sniff.
        """
        self.packet_counts[packet.name] += 1

        while packet.payload:
            packet = packet.payload
            self.packet_counts[packet.name] += 1

        return None

    def interval_handler(self, interval=None):
        """
        Makes a dumpling at regular intervals which lists all the layers seen
        in all the packets so far, and the count for each layer.
        """
        return {'packet_counts': self.packet_counts}



from scapy.all import ARP

from netdumplings import DumplingChef


class ARPChef(DumplingChef):
    """
    Makes dumplings which describe ARP (Address Resolution Protocol) activity.

    Dumpling payload example: ::

        {
            "dst_hw": "11:22:33:44:55:66",
            "dst_ip": "10.0.1.99",
            "notes": "source device is new",
            "operation": "reply",
            "src_hw": "aa:bb:cc:dd:ee:ff",
            "src_ip": "10.0.1.100",
            "time": 1499041214.205803
        }
    """
    request = 1
    reply = 2

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ip_mac = {}

    def packet_handler(self, packet):
        """
        Processes a packet from nd-sniff.  Makes a dumpling summarizing the
        contents of each each valid ARP packet.

        :param packet: Packet from nd-sniff.
        """
        if not packet.haslayer("ARP"):
            return

        arp = packet[ARP]

        if arp.op == ARPChef.request:
            operation = 'request'
        elif arp.op == ARPChef.reply:
            operation = 'reply'
        else:
            operation = arp.op

        payload = {
            'operation': operation,
            'src_hw': arp.hwsrc,
            'src_ip': arp.psrc,
            'dst_hw': arp.hwdst,
            'dst_ip': arp.pdst,
            'time': arp.time,
            'notes': None,
        }

        if arp.op == ARPChef.reply:
            if self.ip_mac.get(arp.psrc) is None:
                payload['notes'] = 'source device is new'
            elif (self.ip_mac.get(arp.psrc) and
                  self.ip_mac[arp.psrc] != arp.hwsrc):
                payload['notes'] = 'source device has new IP address'

            # Remember this IP -> MAC mapping.
            self.ip_mac[arp.psrc] = arp.hwsrc

        return payload



import datetime

import click
import colorama
import termcolor

import netdumplings
from netdumplings._shared import HUB_HOST, HUB_OUT_PORT

from netdumplings.console._shared import CLICK_CONTEXT_SETTINGS


PRINT_COLOR = False


async def on_connect(hub_uri, websocket):
    """
    Called when the connection to ``nd-hub`` has been created.

    :param hub_uri: The ``nd-hub`` websocket URI.
    :param websocket: The websocket object used for talking to ``nd-hub``
        (websockets.WebSocketClientProtocol).
    """
    print('nd-hub status from {0}'.format(hub_uri))
    print('Waiting for data... ', end='', flush=True)


async def on_dumpling(dumpling):
    """
    Called when a new dumpling is received from ``nd-hub``. Prints summary
    information about the current state of ``nd-hub``.

    :param dumpling: The received dumpling.
    """
    payload = dumpling.payload

    up_mins, up_secs = divmod(int(payload['server_uptime']), 60)
    up_hrs, up_mins = divmod(up_mins, 60)
    up_str = '{0:02d}:{1:02d}:{2:02d}'.format(up_hrs, up_mins, up_secs)

    up_str = (
        termcolor.colored(up_str, attrs=['bold']) if PRINT_COLOR else up_str
    )

    dumplings_in = (
        termcolor.colored(payload['total_dumplings_in'], attrs=['bold'])
        if PRINT_COLOR else payload['total_dumplings_in']
    )

    dumplings_out = (
        termcolor.colored(payload['total_dumplings_out'], attrs=['bold'])
        if PRINT_COLOR else payload['total_dumplings_out']
    )

    kitchens = (
        termcolor.colored(payload['dumpling_kitchen_count'], attrs=['bold'])
        if PRINT_COLOR else payload['dumpling_kitchen_count']
    )

    eaters = (
        termcolor.colored(payload['dumpling_eater_count'], attrs=['bold'])
        if PRINT_COLOR else payload['dumpling_eater_count']
    )

    status_msg = (
        '\r{now}  uptime: {uptime}  dumplings in: {dumplings_in}  '
        'out: {dumplings_out}  kitchens: {kitchens}  eaters: {eaters} '.format(
            now=datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
            dumplings_in=dumplings_in,
            dumplings_out=dumplings_out,
            uptime=up_str,
            kitchens=kitchens,
            eaters=eaters,
        )
    )

    print(status_msg, end='', flush=True)


async def on_connection_lost(e):
    """
    Called when the ``nd-hub`` connection is lost.

    :param e: The exception thrown during the connection close.
    """
    print('\nLost connection to nd-hub: {}'.format(e))


# -----------------------------------------------------------------------------

@click.command(
    context_settings=CLICK_CONTEXT_SETTINGS,
)
@click.option(
    '--hub', '-h',
    help='Address where nd-hub is sending dumplings from.',
    metavar='HOST:PORT',
    default='{}:{}'.format(HUB_HOST, HUB_OUT_PORT),
    show_default=True,
)
@click.option(
    '--eater-name', '-n',
    help='Dumpling eater name for this tool when connecting to nd-hub.',
    metavar='EATER_NAME',
    default='statuseater',
    show_default=True,
)
@click.option(
    '--color / --no-color',
    help='Print color output.',
    default=True,
    show_default=True,
)
@click.version_option(version=netdumplings.__version__)
def hubstatus_cli(hub, eater_name, color):
    """
    A dumpling eater.

    Connects to nd-hub (the dumpling hub) and continually prints summary status
    information from any SystemStatusChef dumplings. This is a system
    monitoring dumpling eater which can be used to keep an eye on nd-hub.
    """
    global PRINT_COLOR
    PRINT_COLOR = color

    if PRINT_COLOR:
        colorama.init()  # Needed for Windows console.

    eater = netdumplings.DumplingEater(
        name=eater_name,
        hub=hub,
        chef_filter=['SystemStatusChef'],
        on_connect=on_connect,
        on_dumpling=on_dumpling,
        on_connection_lost=on_connection_lost,
    )

    eater.run()


if __name__ == '__main__':
    hubstatus_cli()



import logging
from typing import Optional

import scapy.packet

import netdumplings  # noqa
from ._shared import JSONSerializable


class DumplingChef:
    """
    Base class for all dumpling chefs.

    **DumplingChef objects are instantiated for you by nd-sniff. You normally
    won't need to instantiate DumplingChef objects yourself. Instead you'll
    normally subclass DumplingChef in a Python module which you'll pass to
    nd-sniff on the commandline.**

    When instantiated, a DumplingChef registers itself with the given
    ``kitchen`` which will then take care of calling the chef's packet and
    interval handlers as appropriate.

    This class implements the following methods which will usually be
    overridden by subclasses:

    * :meth:`packet_handler`
    * :meth:`interval_handler`

    :param kitchen: The dumpling kitchen which is providing the network packets
        used to create the dumplings.
    """
    # Setting assignable_to_kitchen to False (in a subclass) will ensure the
    # chef cannot be assigned to any kitchens via nd-sniff.
    assignable_to_kitchen = True

    def __init__(
            self,
            kitchen: Optional['netdumplings.DumplingKitchen'] = None,
    ) -> None:
        """
        """
        self.kitchen = kitchen
        self.name = type(self).__name__
        self.dumplings_sent_count = 0
        self._logger = logging.getLogger(__name__)

        if self.kitchen:
            self.kitchen.register_chef(self)

    def __repr__(self):
        return '{}(kitchen={})'.format(type(self).__name__, repr(self.kitchen))

    def packet_handler(self, packet: scapy.packet.Raw) -> JSONSerializable:
        """
        Called automatically by the dumpling kitchen (``nd-sniff``) whenever a
        new packet has been sniffed.

        This method is expected to be overridden by child classes. This base
        implementation returns a payload which is the string representation of
        the packet.

        The return value is turned into a dumpling by the kitchen. If ``None``
        is returned then no dumpling will be created.

        :param packet: Network packet (from scapy).
        :rtype: Anything which is JSON-serializable.
        :return: Dumpling payload.
        """
        payload = "{0}: {1}".format(type(self).__name__, packet.summary())
        self._logger.debug("{0}: Received packet: {1}",
                           self.name, packet.summary())

        return payload

    def interval_handler(
            self, interval: Optional[int] = None) -> JSONSerializable:
        """
        Called automatically at regular intervals by the dumpling kitchen
        (``nd-sniff``).

        Allows for time-based (rather than purely packet-based) chefs to keep
        on cheffing even in the absence of fresh packets.

        This method is expected to be overridden by child classes. This base
        implementation does nothing but log a debug entry.

        The return value is turned into a dumpling by the kitchen. If ``None``
        is returned then no dumpling will be created.

        :param interval: Frequency (in secs) of the time interval pokes.
        :rtype: Anything which is JSON-serializable.
        :return: Dumpling payload.
        """
        self._logger.debug(
            "{0}: Received interval_handler poke", self.name)

        return None



import importlib
import importlib.util
import inspect
import logging
import multiprocessing
import os
import os.path
import sys
from threading import Thread
from time import sleep
from typing import Dict, List, Optional

from scapy.all import sniff
from scapy.error import Scapy_Exception
import scapy.packet

from .dumpling import Dumpling, DumplingDriver
from .dumplingchef import DumplingChef
from ._shared import JSONSerializable


class DumplingKitchen:
    """
    A network packet sniffer kitchen.

    The sniffer kitchen does the following:

    * Sniffs network packets
    * Allows chefs to be registered with the kitchen
    * Sends all sniffed packets to all registered dumpling chefs' packet
      handlers
    * Optionally pokes the interval handler of all registered dumpling chefs
      at regular time intervals
    * Takes the return values from the packet and interval handlers, converts
      them into dumplings, JSON-serializes the dumplings, and puts them on
      the given dumpling queue

    Once the kitchen has put a dumpling on the dumpling queue, the kitchen's
    involvement in that dumpling's life cycle is complete. It's the
    responsibility of the thing instantiating the kitchen to provide the queue
    and to then pull dumplings off the queue and send them on to the dumpling
    hub.

    Which network packets are sniffed can be controlled with a PCAP-style
    packer filter.

    :param dumpling_queue: Queue for sending dumplings to the dumpling hub.
    :param name: Kitchen name.
    :param interface: Network interface to sniff on.
    :param sniffer_filter: PCAP-compliant sniffer filter (``None`` means sniff
        all packets).
    :param chef_poke_interval: Frequency (in secs) to call all registered chef
        poke handlers. ``None`` disables poking.
    """
    def __init__(
            self,
            dumpling_queue: multiprocessing.Queue,
            name: str = 'default',
            interface: str = 'all',
            sniffer_filter: str = 'tcp',
            chef_poke_interval: int = 5,
    ) -> None:
        self.name = name
        self.interface = interface
        self.filter = sniffer_filter
        self.chef_poke_interval = chef_poke_interval
        self.dumpling_queue = dumpling_queue

        self._chefs = []
        self._logger = logging.getLogger(__name__)

    def __repr__(self):
        return (
            '{}('
            'dumpling_queue={}, '
            'name={}, '
            'interface={}, '
            'sniffer_filter={}, '
            'chef_poke_interval={})'.format(
                type(self).__name__,
                repr(self.dumpling_queue),
                repr(self.name),
                repr(self.interface),
                repr(self.filter),
                repr(self.chef_poke_interval),
            )
        )

    def _put_dumpling_on_queue(
            self,
            chef: DumplingChef,
            payload: JSONSerializable,
            driver: DumplingDriver,
    ):
        """
        Creates a dumpling, JSON-serializes it, and puts it on the dumpling
        queue.

        :param chef: The chef which provided the dumpling payload.
        :param payload: The dumpling payload.
        :param driver: The driver of the dumpling creation.
        """
        dumpling = Dumpling(chef=chef, payload=payload, driver=driver)
        dumpling_json = dumpling.to_json()
        self.dumpling_queue.put(dumpling_json)
        self._logger.debug("{0}: Put dumpling on queue, {1} bytes".format(
            self.name, len(dumpling_json)
        ))

    def _process_packet(self, packet: scapy.packet.Raw):
        """
        Takes a single sniffed packet and:

        * Passes it to each of the registered dumpling chef packet handlers
        * Takes the returned payload from each chef (if any) and initiates the
          process of putting the result on the dumpling queue

        If a packet handler raises an exception then the exception will be
        logged and otherwise ignored.

        :param packet: The network packet to process (from scapy).
        """
        for chef in self._chefs:
            try:
                payload = chef.packet_handler(packet)
            except Exception as e:
                self._logger.exception(
                    "{0}: Error invoking packet handler for chef {1}: "
                    "{2}".format(self.name, chef.name, e)
                )

                continue

            if payload is not None:
                self._put_dumpling_on_queue(
                    chef, payload, DumplingDriver.packet
                )

    def _poke_chefs(self, interval: int):
        """
        Call any registered dumpling chef interval handlers at regular time
        intervals.

        This is intended to be run in a separate thread.

        If an interval handler raises an exception then the exception will be
        logged and otherwise ignored.

        :param interval: Frequency (in secs) of calls to the interval handlers.
        """
        while True:
            self._logger.debug("{0}: Poking chefs".format(self.name))

            for chef in self._chefs:
                try:
                    payload = chef.interval_handler(interval=interval)
                except Exception as e:
                    self._logger.exception(
                        "{0}: Error invoking interval handler for chef {1}: "
                        "{2}".format(self.name, chef.name, e)
                    )

                    continue

                if payload is not None:
                    self._put_dumpling_on_queue(
                        chef, payload, DumplingDriver.interval
                    )

            sleep(interval)

    @staticmethod
    def get_chefs_in_modules(chef_modules: Optional[List[str]] = None) -> Dict:
        """
        Finds available :class:`DumplingChef` subclasses.

        This helper method looks inside the modules provided in
        ``chef_modules`` for classes which are subclasses of
        :class:`DumplingChef`.

        The returned dict contains information on all found chefs. Each key
        is the name of a single module, and the values are dicts which contain
        the follownig keys:

        * ``"chef_classes"`` - a list of DumplingChef class names contained in
          the module
        * ``"import_error"`` - an error string describing a problem encountered
          while finding dumpling chefs in the module (``False`` if no errors)

        :return: Information on chefs found in the give modules.
        """
        # Allow for a chef module to be relative to the current working
        # directory (wherever the script calling this method is being run
        # from). There's potential for this to result in unexpected behaviour
        # as we're effectively modifying the PYTHONPATH.
        sys.path.append(os.getcwd())

        chef_info = {}

        for chef_module in chef_modules:
            is_py_file = True if os.path.isfile(chef_module) else False

            chef_info[chef_module] = {
                'import_error': False,
                'chef_classes': [],
                'is_py_file': is_py_file,
            }

            # Import the module for subsequent Chef extraction.
            if is_py_file:
                try:
                    spec = importlib.util.spec_from_file_location(
                        'chefs', chef_module
                    )
                    module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(module)
                except AttributeError:
                    # Non-Python files result in spec not having a loader attr.
                    chef_info[chef_module]['import_error'] = (
                        'does not appear to be an importable Python file'
                    )
                    continue
            else:
                try:
                    module = importlib.import_module(chef_module)
                except ImportError as e:
                    chef_info[chef_module]['import_error'] = str(e)
                    continue

            chef_classes = inspect.getmembers(module, inspect.isclass)

            for chef_class in chef_classes:
                if chef_class[0] == 'DumplingChef':
                    continue

                if issubclass(chef_class[1], DumplingChef):
                    chef_info[chef_module]['chef_classes'].append(
                        chef_class[0]
                    )

        return chef_info

    def register_chef(self, chef: DumplingChef):
        """
        Registers dumpling chef with the kitchen.

        Registered dumpling chefs will have their packet and interval handlers
        invoked by the kitchen as appropriate.

        :param chef: The DumplingChef to register.
        """
        self._chefs.append(chef)

        self._logger.debug("{0}: Received chef registration from {1}".format(
            self.name, chef.name
        ))

    def run(self):
        """
        Starts the kitchen.

        This blocks and will run forever.
        """
        # Start the chef poking thread.
        if self.chef_poke_interval is not None:
            self._logger.info(
                "{0}: Starting interval poker thread".format(self.name))

            interval_poker = Thread(
                target=self._poke_chefs,
                kwargs={'interval': self.chef_poke_interval},
                daemon=True,
            )

            interval_poker.start()
        else:
            self._logger.info(f"{self.name}: Interval poker thread disabled")

        # Start the sniffer.
        self._logger.info(f"{self.name}: Sniffing started")

        try:
            if self.interface == 'all':
                sniff(filter=self.filter, prn=self._process_packet, store=0)
            else:
                sniff(iface=self.interface, filter=self.filter,
                      prn=self._process_packet, store=0)
        except Scapy_Exception as e:
            self._logger.error(
                "The sniffer encountered a problem (it might help to run as "
                "root if you're not already)"
            )
            self._logger.error(f"Error from scapy: {e}")
        except ValueError as e:
            # On Windows an unknown interface is a ValueError (on OS X it's
            # picked up by Scapy_Exception).
            if "Unknown pypcap network interface" in str(e):
                self._logger.error(str(e))
            else:
                raise



import json
import logging
import logging.config
import os
import time
from typing import Dict, List, Union

from .exceptions import InvalidDumpling


# For typing.
JSONSerializable = Union[str, int, List, Dict, None]

# Tuples of (code, msg) for sending when closing websocket connections.
ND_CLOSE_MSGS = {
    'conn_cancelled': (4101, "connection cancelled"),
    'eater_full': (4102, "dumpling eater is full"),
}

LOGGING_CONFIG_FILE = os.path.join(
    os.path.dirname(__file__), 'data', 'logging.json'
)

HUB_HOST = 'localhost'
HUB_IN_PORT = 11347
HUB_OUT_PORT = 11348
HUB_STATUS_FREQ = 5


def configure_logging(log_level=logging.INFO, config_file=LOGGING_CONFIG_FILE):
    """
    Configure logging. Configuration is retrieved from an external
    file (``data/logging.json`` in the netdumplings package), which can be
    overridden with ``config_file`` or with the `NETDUMPLINGS_LOGGING_CONFIG``
    environment variable.

    :param log_level: Log level to use. Defaults to ``'INFO'``.
    :param config_file: Path to the logging config file to use. Defaults to
        ``data/logging.json`` in the netdumplings package.
    """
    # Format timestamps in GMT, YYYY-MM-DDThh:mm:ss.sss
    logging.Formatter.converter = time.gmtime
    logging.Formatter.default_time_format = '%Y-%m-%dT%H:%M:%S'
    logging.Formatter.default_msec_format = '%s.%03d'

    use_basic_config = False

    try:
        config_file = os.environ['NETDUMPLINGS_LOGGING_CONFIG']
    except KeyError:
        pass

    if os.path.exists(config_file):
        try:
            with open(config_file) as logging_config_handler:
                logging_config = json.load(logging_config_handler)

            logging.config.dictConfig(logging_config)
        except (IOError, json.decoder.JSONDecodeError) as e:
            use_basic_config = True
            print(
                'error loading logging config: {}: {}'.format(config_file, e)
            )
    else:
        use_basic_config = True

    if use_basic_config:
        logging.basicConfig(level=log_level)


def validate_dumpling(dumpling_json):
    """
    Validates a dumpling received from (or about to be sent to) the dumpling
    hub. Validation involves ensuring that it's valid JSON and that it includes
    a ``metadata.chef`` key.

    :param dumpling_json: The dumpling JSON.
    :raise: :class:`netdumplings.exceptions.InvalidDumpling` if the
        dumpling is invalid.
    :return: A dict created from the dumpling JSON.
    """
    try:
        dumpling = json.loads(dumpling_json)
    except json.JSONDecodeError as e:
        raise InvalidDumpling("Could not interpret dumpling JSON")

    try:
        dumpling['metadata']['chef']
    except (KeyError, TypeError) as e:
        raise InvalidDumpling("Could not determine chef name")

    return dumpling



import json

import pygments
import pygments.lexers
import pygments.formatters


CLICK_CONTEXT_SETTINGS = {
    'max_content_width': 100,
    'help_option_names': ['--help'],
}


def printable_dumpling(contents, colorize=True):
    """
    Converts dumpling contents to a pretty-printed and optionally colorized
    output string.

    :param contents: Dumpling contents to generate printable string from.
    :param colorize: Whether to colorize the dumpling string.
    :return: Pretty-printed string representing the dumpling contents.
    """
    contents_printable = json.dumps(
        contents,
        sort_keys=True,
        indent=4,
        separators=(',', ': ')
    )

    if colorize:
        contents_printable = pygments.highlight(
            contents_printable,
            pygments.lexers.JsonLexer(),
            pygments.formatters.TerminalFormatter(),
        ).rstrip()

    return contents_printable



import os
from setuptools import setup


here = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(os.path.join(here, 'netdumplings', '_version.py')) as ver_file:
    exec(ver_file.read(), version)

with open(os.path.join(here, 'README.rst'), 'r') as f:
    readme = f.read()

packages = [
    'netdumplings',
    'netdumplings.console',
    'netdumplings.dumplingchefs',
]

install_requires = [
    'click~=7.1',
    'colorama',
    'pygments',
    'scapy~=2.4.3',
    'termcolor',
    'websockets~=8.1.0',
]

tests_require = [
    'asynctest',
    'pytest',
    'pytest-asyncio',
    'pytest-cov',
    'pytest-mock',
    'pytest-sugar',
]

extras_require = {
    'dev': [
        'flake8',
        'mypy',
        'sphinx',
        'sphinx-autodoc-typehints',
    ] + tests_require,
}

setup(
    name='netdumplings',
    version=version['__version__'],
    description=(
        'A framework for distributed network packet sniffing and processing.'
    ),
    long_description=readme,
    author='Mike Joblin',
    author_email='mike@tastymoss.com',
    url='https://github.com/mjoblin/netdumplings',
    packages=packages,
    include_package_data=True,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
    download_url='https://pypi.python.org/pypi/netdumplings',
    entry_points={
        'console_scripts': [
            'nd-hub=netdumplings.console:hub_cli',
            'nd-hubdetails=netdumplings.console:hubdetails_cli',
            'nd-hubstatus=netdumplings.console:hubstatus_cli',
            'nd-print=netdumplings.console:print_cli',
            'nd-sniff=netdumplings.console:sniff_cli',
        ]
    },
    license='MIT',
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Networking',
        'Topic :: System :: Networking :: Monitoring',
    ]
)



import click

import netdumplings
from netdumplings._shared import HUB_HOST, HUB_OUT_PORT

from netdumplings.console._shared import CLICK_CONTEXT_SETTINGS, printable_dumpling


PRINT_COLOR = False


async def on_connect(hub_uri, websocket):
    """
    Called when the connection to ``nd-hub`` has been created.

    :param hub_uri: The ``nd-hub`` websocket URI.
    :param websocket: The websocket object used for talking to nd-hub
        (websockets.WebSocketClientProtocol).
    """
    print('Connected to nd-hub at {0}'.format(hub_uri))
    print('Waiting for a SystemStatus dumpling...')


async def on_dumpling(dumpling):
    """
    Called when a new dumpling is received from ``nd-hub``. Prints dumpling
    payload.

    :param dumpling: The received dumpling.
    """
    print('\n{}\n'.format(
        printable_dumpling(dumpling.payload, colorize=PRINT_COLOR)
    ))


async def on_connection_lost(e):
    """
    Called when the ``nd-hub`` connection is lost.

    :param e: The exception thrown during the connection close.
    """
    print('\nLost connection to nd-hub: {}'.format(e))


# -----------------------------------------------------------------------------

@click.command(
    context_settings=CLICK_CONTEXT_SETTINGS,
)
@click.option(
    '--hub', '-h',
    help='Address where nd-hub is sending dumplings from.',
    metavar='HOST:PORT',
    default='{}:{}'.format(HUB_HOST, HUB_OUT_PORT),
    show_default=True,
)
@click.option(
    '--eater-name', '-n',
    help='Dumpling eater name for this tool when connecting to nd-hub.',
    metavar='EATER_NAME',
    default='detailseater',
    show_default=True,
)
@click.option(
    '--color / --no-color',
    help='Print color output.',
    default=True,
    show_default=True,
)
@click.version_option(version=netdumplings.__version__)
def hubdetails_cli(hub, eater_name, color):
    """
    A dumpling eater.

    Connects to nd-hub (the dumpling hub) and waits for a single SystemStatus
    dumpling which it displays the full contents of and then exits.
    """
    global PRINT_COLOR
    PRINT_COLOR = color

    eater = netdumplings.DumplingEater(
        name=eater_name,
        hub=hub,
        chef_filter=['SystemStatusChef'],
        on_connect=on_connect,
        on_dumpling=on_dumpling,
        on_connection_lost=on_connection_lost,
    )

    eater.run(dumpling_count=1)


# -----------------------------------------------------------------------------

if __name__ == '__main__':
    hubdetails_cli()
